import datetime
import re
from typing import Optional, Tuple

from pydantic import BaseModel, Field, ValidationError, field_validator

from avtdl.plugins.youtube.common import extract_keys, find_all, find_one, get_continuation_token


class VideoRendererInfo(BaseModel):
    video_id: str
    url: str
    title: str
    summary: Optional[str] = Field(repr=False)
    scheduled: Optional[datetime.datetime] = None
    author: Optional[str]
    avatar_url: Optional[str] = None
    channel_link: Optional[str] = None
    channel_id: Optional[str] = None
    published_text: Optional[str]
    length: Optional[str]

    is_upcoming: bool
    is_live: bool
    is_member_only: bool

def get_video_renderers(page: str, anchor: str = 'var ytInitialData = ') -> Tuple[list, Optional[str], dict]:
    keys = ['gridVideoRenderer', 'videoRenderer', 'playlistVideoRenderer', 'continuationEndpoint']
    items, data = extract_keys(page, keys, anchor)
    continuation_token = get_continuation_token(items.pop('continuationEndpoint', {}))
    renderers = []
    for item in items.values():
        renderers.extend(item)
    return renderers, continuation_token, data

def parse_scheduled(timestamp: Optional[str]) -> Optional[datetime.datetime]:
    if timestamp is None:
        return None
    try:
        timestamp_value = float(timestamp)
    except ValueError:
        return None
    scheduled = datetime.datetime.fromtimestamp(timestamp_value, tz=datetime.timezone.utc)
    return scheduled

def get_author_fallback(item:dict) -> Optional[str]:
    full_text = find_one(item, '$.title.accessibility..label')
    title_part = find_one(item, '$.title.runs..text')
    views_part = find_one(item, '$.viewCountText.simpleText')
    views_part_fallback = re.search('\d+\D+$', full_text) if full_text else None
    if not all((full_text, title_part, views_part or views_part_fallback)):
        return None
    start = full_text.find(title_part)

    if views_part is not None:
        end = full_text.find(views_part)
    elif views_part_fallback is not None:
        end = views_part_fallback.start()
    else:
        end = -1

    if -1 in [start, end]:
        return None
    try:
        author_text = full_text[start + len(title_part) : end].split(':')[1].strip()
    except (IndexError, AttributeError):
        return None
    return author_text


class AuthorInfo(BaseModel):
    name: str
    channel: str
    channel_id: str
    avatar_url: Optional[str] = None
    
    @field_validator('channel')
    @classmethod
    def add_prefix(cls, channel: str) -> str:
        if channel.startswith('/'):
            return 'https://www.youtube.com' + channel
        else:
            return channel


def parse_author(video_render: dict) -> Optional[AuthorInfo]:
    author_info = find_one(video_render, '$.ownerText,shortBylineText')
    if author_info is None:
        return None
    author = find_one(author_info, '$..text')
    channel_link = find_one(author_info, '$..browseEndpoint.canonicalBaseUrl')
    channel_id = find_one(author_info, '$..browseEndpoint.browseId')

    avatar_url = find_one(video_render, '$.channelThumbnailSupportedRenderers..thumbnails.[::-1].url')

    try:
        return AuthorInfo(name=author, channel=channel_link, channel_id=channel_id, avatar_url=avatar_url)
    except ValidationError:
        return None

def parse_owner_info(page: dict) -> Optional[AuthorInfo]:
    owner_info_data = find_one(page, '$.header.c4TabbedHeaderRenderer')
    if owner_info_data is None:
        return None
    author = find_one(owner_info_data, '$.title')
    channel_link = find_one(owner_info_data, '$.navigationEndpoint.browseEndpoint.canonicalBaseUrl')
    channel_id = find_one(owner_info_data, '$.channelId')
    avatar_url = find_one(owner_info_data, '$.avatar.thumbnails.[::-1].url')
    try:
        return AuthorInfo(name=author, channel=channel_link, channel_id=channel_id, avatar_url=avatar_url)
    except ValidationError:
        return None

def parse_video_renderer(item: dict, owner_info: Optional[AuthorInfo], raise_on_error: bool = False) -> Optional[VideoRendererInfo]:
    video_id = item.get('videoId')
    url = f'https://www.youtube.com/watch?v={video_id}'
    title = find_one(item, '$.title..text,simpleText')
    summary = find_one(item, '$.descriptionSnippet..text')
    
    author_info = parse_author(item) or owner_info
    if author_info is None:
        author_name = get_author_fallback(item) 
        channel_link = channel_id = avatar_url = None
    else:
        author_name = author_info.name
        channel_link = author_info.channel
        channel_id = author_info.channel_id
        avatar_url = author_info.avatar_url

    scheduled_timestamp = find_one(item, '$.upcomingEventData.startTime')
    scheduled = parse_scheduled(scheduled_timestamp)
    published_text = find_one(item, '$.publishedTimeText.simpleText')
    length = find_one(item, '$.lengthText.simpleText') or find_one(item, '$..thumbnailOverlayTimeStatusRenderer.simpleText')

    badges = find_all(item, '$.badges..style')
    is_member_only = 'BADGE_STYLE_TYPE_MEMBERS_ONLY' in badges
    is_live = 'BADGE_STYLE_TYPE_LIVE_NOW' in badges
    is_upcoming = scheduled is not None

    try:
        info = VideoRendererInfo(video_id=video_id,
                                 url=url,
                                 title=title,
                                 summary=summary,
                                 scheduled=scheduled,
                                 author=author_name,
                                 avatar_url=avatar_url,
                                 channel_link=channel_link,
                                 channel_id=channel_id,
                                 published_text=published_text,
                                 length=length,
                                 is_live=is_live,
                                 is_upcoming=is_upcoming,
                                 is_member_only=is_member_only)
        return info
    except ValidationError:
        if raise_on_error:
            raise
        return None

def handle_page(page: str) -> list:
    items, continuation, data = get_video_renderers(page)
    owner_info = parse_owner_info(data)
    info = [parse_video_renderer(x, owner_info) for x in items]
    return info


import datetime
import json
from json import JSONDecodeError
from typing import Any, List, Optional, Sequence, Tuple

import aiohttp
from pydantic import Field, ValidationError

from avtdl.core import utils
from avtdl.core.interfaces import Filter, FilterEntity, Record
from avtdl.core.monitors import PagedFeedMonitor, PagedFeedMonitorConfig, PagedFeedMonitorEntity
from avtdl.core.plugins import Plugins
from avtdl.plugins.filters.filters import EmptyFilterConfig
from avtdl.plugins.youtube.common import get_innertube_context, handle_consent, prepare_next_page_request, thumbnail_url
from avtdl.plugins.youtube.feed_info import AuthorInfo, VideoRendererInfo, get_video_renderers, parse_owner_info, parse_video_renderer


@Plugins.register('channel', Plugins.kind.ASSOCIATED_RECORD)
@Plugins.register('filter.channel', Plugins.kind.ASSOCIATED_RECORD)
class YoutubeVideoRecord(VideoRendererInfo, Record):
    """
    Youtube video or livestream listed among others on Youtube page

    Produced by parsing a channels main page, videos and streams tab,
    as well as playlists, and, with login cookies, subscriptions feed.
    """

    video_id: str
    """short string identifying video on Youtube. Part of video url"""
    url: str
    """link to the video, uses `https://www.youtube.com/watch?v=<video_id>` format"""
    title: str
    """title of the video at the time of parsing"""
    summary: Optional[str] = Field(repr=False)
    """snippet of the video description. Not always available"""
    scheduled: Optional[datetime.datetime] = None
    """scheduled date for upcoming stream or premiere"""
    author: Optional[str]
    """channel name"""
    avatar_url: Optional[str] = None
    """link to the avatar of the channel. Not always available"""
    channel_link: Optional[str] = None
    """link to the channel uploading the video"""
    channel_id: Optional[str] = None
    """channel ID in old format (such as `UCK0V3b23uJyU4N8eR_BR0QA`)"""
    published_text: Optional[str]
    """localized text saying how long ago the video was uploaded"""
    length: Optional[str]
    """text showing the video duration (hh:mm:ss)"""

    is_upcoming: bool
    """indicates that video is an upcoming livestream or premiere"""
    is_live: bool
    """indicates that the video is a livestream or premiere that is currently live"""
    is_member_only: bool
    """indicated that the video is limited to members of the channel. Note that the video status might be changed at any time"""

    def __str__(self):
        scheduled = self.scheduled
        if scheduled:
            scheduled_time = '\nscheduled to {}'.format(scheduled.strftime('%Y-%m-%d %H:%M'))
        else:
            scheduled_time = ''
        template = '{}\n{}\npublished by {}'
        return template.format(self.url, self.title, self.author) + scheduled_time

    def __repr__(self):
        template = '{:<8} [{}] {}'
        return template.format(self.author or 'Unknown author', self.video_id, self.title[:60])

    def discord_embed(self) -> dict:
        embed = {
            'title': self.title,
            # 'description': ,
            'url': self.url,
            'color': None,
            'author': {'name': self.author, 'url': self.channel_link, 'icon_url': self.avatar_url},
            'image': {'url': thumbnail_url(self.video_id)},
            'fields': []
        }
        footer = ''
        if self.published_text:
            footer += self.published_text
        if self.length:
            embed['fields'].append({'name': f'[{self.length}]', 'value': '', 'inline': True})
        if self.scheduled is not None:
            scheduled = self.scheduled.strftime('%Y-%m-%d %H:%M')
            embed['fields'].append({'name': 'Scheduled:', 'value': scheduled, 'inline': True})
        embed['footer'] = {'text': footer}
        return embed


@Plugins.register('channel', Plugins.kind.ACTOR_CONFIG)
class VideosMonitorConfig(PagedFeedMonitorConfig):
    pass


@Plugins.register('channel', Plugins.kind.ACTOR_ENTITY)
class VideosMonitorEntity(PagedFeedMonitorEntity):
    update_interval: float = 1800


@Plugins.register('channel', Plugins.kind.ACTOR)
class VideosMonitor(PagedFeedMonitor):
    """
    Youtube channel monitor

    Monitors Youtube url listing videos, such as channels main page,
    videos and streams tab of a channel, as well as playlists, and,
    with login cookies, subscriptions feed or the main page.

    Due to small differences in presentation in aforementioned
    sources, same video might have slightly different appearance when
    parsed from different urls. For example, video parsed from main
    page or subscriptions feed will not have full description text.

    Examples of supported url:

    - `https://www.youtube.com/@ChannelName`
    - `https://www.youtube.com/@ChannelName/videos`
    - `https://www.youtube.com/@ChannelName/streams`
    - `https://www.youtube.com/channel/UCK0V3b23uJyU4N8eR_BR0QA/`
    - `https://www.youtube.com/playlist?list=PLWGY3fcU-ZeQmBfoJ6SmT8v2zV8NEhrB2`
    - `https://www.youtube.com/feed/subscriptions` (providing cookies is necessarily)

    Unlike `rss` monitor, with login cookies it can see videos and streams
    with limited access (such as member-only).

    While monitoring a single channel is less efficient, using this monitor with
    subscriptions feed url on a dedicated account is a recommended way
    to monitor a high amount (hundreds) of channels, as it only requires
    loading a single page to check all of them for updates.
    """

    async def handle_first_page(self, entity: PagedFeedMonitorEntity, session: aiohttp.ClientSession) -> Tuple[Optional[Sequence[Record]], Optional[Any]]:
        raw_page_text = await self.request(entity.url, entity, session)
        if raw_page_text is None:
            return None, None
        raw_page_text = await handle_consent(raw_page_text, entity.url, session, self.logger)
        video_renderers, continuation_token, page = get_video_renderers(raw_page_text)
        owner_info = parse_owner_info(page)
        current_page_records = self._parse_entries(owner_info, video_renderers, entity)
        innertube_context = get_innertube_context(raw_page_text)
        return current_page_records, (innertube_context, owner_info, continuation_token)

    async def handle_next_page(self, entity: PagedFeedMonitorEntity, session: aiohttp.ClientSession, context: Optional[Any]) -> Tuple[Optional[Sequence[Record]], Optional[Any]]:
        innertube_context, owner_info, continuation_token = context  # type: ignore
        if continuation_token is None:
            self.logger.debug(f'[{entity.name}] no continuation for next page, done loading')
            return [], None

        url, headers, post_body = prepare_next_page_request(innertube_context, continuation_token, cookies=session.cookie_jar)
        raw_page = await utils.request(url, session, self.logger, method='POST', headers=headers,
                                       data=json.dumps(post_body), retry_times=3, retry_multiplier=2,
                                       retry_delay=5)
        if raw_page is None:
            self.logger.debug(f'[{entity.name}] failed to load next page, aborting')
            return None, None
        video_renderers, continuation_token, page = get_video_renderers(raw_page, anchor='')
        context = (innertube_context, owner_info, continuation_token) if continuation_token else None
        current_page_records = self._parse_entries(owner_info, video_renderers, entity)
        return current_page_records, context

    def _parse_entries(self, owner_info: Optional[AuthorInfo], video_renderers: List[dict], entity: PagedFeedMonitorEntity) -> List[YoutubeVideoRecord]:
        records: List[YoutubeVideoRecord] = []
        for item in video_renderers:
            try:
                info = parse_video_renderer(item, owner_info, raise_on_error=True)
                record = YoutubeVideoRecord.model_validate(info.model_dump())
                records.append(record)
            except (ValueError, JSONDecodeError, ValidationError) as e:
                self.logger.warning(f'[{entity.name}] failed to parse video renderer on "{entity.url}": {type(e)}: {e}')
                self.logger.debug(f'[{entity.name}] raw video renderer:\n{item}')
                continue
        if not records:
            self.logger.warning(f'[{entity.name}] parsing page "{entity.url}" yielded no videos, check url and cookies')
        records = records[::-1] # records are ordered from old to new on page, reorder in chronological order
        return records

    def get_record_id(self, record: YoutubeVideoRecord) -> str:
        return record.video_id


@Plugins.register('filter.channel', Plugins.kind.ACTOR_CONFIG)
class ChannelFilterConfig(EmptyFilterConfig):
    pass

@Plugins.register('filter.channel', Plugins.kind.ACTOR_ENTITY)
class ChannelFilterEntity(FilterEntity):
    upcoming: bool = False
    """to pass the filter a record should be either upcoming livestream or scheduled premiere"""
    live: bool = False
    """to pass the filter a record should be an ongoing livestream"""
    member_only: bool = False
    """to pass the filter a record should be marked as member-only"""


@Plugins.register('filter.channel', Plugins.kind.ACTOR)
class ChannelFilter(Filter):
    """
    Pick `YoutubeVideoRecord` with specified properties

    Filter that only lets `YoutubeVideoRecord` through if it has certain properties.
    All records from other sources pass through without filtering.

    If multiple settings are set to `true`, they all should match. Use multiple
    entities if picking records with any of multiple properties is required.
    """

    def __init__(self, config: ChannelFilterConfig, entities: Sequence[ChannelFilterEntity]):
        super().__init__(config, entities)

    def match(self, entity: ChannelFilterEntity, record: YoutubeVideoRecord) -> Optional[YoutubeVideoRecord]:
        if not isinstance(record, YoutubeVideoRecord):
            self.logger.debug(f'[{entity.name}] record is not a YoutubeVideoRecord, letting it through: {record!r}')
            return record
        if entity.upcoming and not record.is_upcoming:
            return None
        if entity.live and not record.is_live:
            return None
        if entity.member_only and not record.is_member_only:
            return None
        return record



import logging
import logging.handlers
from enum import Enum
from pathlib import Path
from typing import Dict, List, Union

from avtdl.core.utils import check_dir


class LogLevel(str, Enum):
    debug = 'DEBUG'
    info = 'INFO'
    warning = 'WARNING'
    error = 'ERROR'


def set_logging_format(level):
    log_format = '%(asctime)s.%(msecs)03d [%(levelname)-7s] [%(name)s] %(message)s'
    datefmt = '%Y/%m/%d %H:%M:%S'
    handler = logging.StreamHandler()
    handler.setLevel(level)
    formatter = logging.Formatter(log_format, datefmt)
    handler.setFormatter(formatter)
    logging.getLogger().addHandler(handler)
    
    logging.getLogger().name = 'avtdl'
    logging.getLogger().setLevel(logging.NOTSET)

def set_file_logger(path: Path, name: str = 'avtdl.log', max_size=1000000, level: LogLevel = LogLevel.debug):
    check_dir(path, create=True)
    path /= name
    log_format = '%(asctime)s.%(msecs)03d [%(levelname)s] [%(name)s] %(message)s'
    datefmt = '%Y/%m/%d %H:%M:%S'
    formatter = logging.Formatter(log_format, datefmt)
    try:
        handler = logging.handlers.RotatingFileHandler(path, maxBytes=max_size, backupCount=10, encoding='utf8')
    except Exception as e:
        logging.error(f'writing log to {path.absolute()} failed: {e}')
        return
    handler.setFormatter(formatter)
    handler.setLevel(getattr(logging, level))
    logging.getLogger().addHandler(handler)
    logging.info(f'writing verbose log to file {path.absolute()}')

def set_logger(name: str, level: Union[LogLevel, int], propagate=True):
    if isinstance(level, LogLevel):
        # since LogLevel itself should only contain valid log level names
        # it should never raise AttributeError, but if it does let it crash
        # so it can get noticed
        level = getattr(logging, level)
    logger = logging.getLogger(name)
    logger.propagate = propagate
    logger.setLevel(level)

def silence_library_loggers():
    set_logger('asyncio', logging.WARNING)
    set_logger('charset_normalizer', logging.WARNING)
    set_logger('slixmpp', logging.ERROR)

class LogFilter(logging.Filter):

    def __init__(self, names: List):
        super().__init__()
        self.names = names

    def filter(self, record: logging.LogRecord):
        for name in self.names:
            if record.name.startswith(name):
                return False
        return True

def override_loglevel(loggers: Dict[str, LogLevel]):
        for name, level in loggers.items():
            set_logger(name, level, propagate=True)



import asyncio
import logging
from dataclasses import dataclass
import sys

import aioxmpp

ON_ERROR_RETRY_DELAY = 60


def instantiate_loggers(names, level):
    '''
    Make instance of logger with specific name and loglevel
    before it gets done by someone else. Used to make
    logging.getLogger(name) call return instance with
    loglevel different from default.

    This script relies on logging.basicConfig() to set default
    logger level and format, so there is no need to pass
    these settings between modules.

    While aioxmpp.Client() accepts `logger` argument, allowing
    to specify loglevel, some underlying modules and libraries
    just create loggers by themself, using default loglevel
    set by logging.basicConfig().

    This leads to debug messages from aioxmpp modules being
    produced when loglevel is set to DEBUG for this script.
    '''
    for name in names:
        logger = logging.getLogger(name)
        logger.setLevel(level)


@dataclass
class Line:
    recepient: str
    message: str

class MSG2JBR:

    @staticmethod
    async def asend(msg, user, passwd, recepient):
        '''send string or list of strings msg to recepient using user/passwd as credentials'''
        if isinstance(msg, str):
            msg = [msg, ]
        if not msg:
            return
        client = aioxmpp.PresenceManagedClient(aioxmpp.JID.fromstr(user), aioxmpp.make_security_layer(passwd))
        async with client.connected():
            for line in msg:
                message = aioxmpp.Message(to=aioxmpp.JID.fromstr(recepient), type_=aioxmpp.MessageType.CHAT)
                message.body[None] = line
                await client.send(message)

    @staticmethod
    def send(msg, user, passwd, recepient):
        '''send blocking'''
        asyncio.get_event_loop().run_until_complete(MSG2JBR.asend(msg, user, passwd, recepient))

    def __init__(self, username, passwd, logger=None):
        self.user = username
        self.passwd = passwd
        self.send_query = []
        self.can_send = (username is not None and passwd is not None and 'aioxmpp' in sys.modules)
        instantiate_loggers(['aioopenssl', 'aiosasl', 'aioxmpp'], logging.ERROR)
        self.logger = logger or logging.getLogger('msg2jbr.aioxmpp')

    def to_be_send(self, recepient, message):
        line = Line(recepient, message)
        self.send_query.append(line)

    async def asend_pending(self):
        if not self.can_send:
            for line in self.send_query:
                warning = 'jabber module required but unable to send message to {}: {}'
                self.logger.debug(warning.format(line.recepient, line.message))
            self.send_query = []
            return
        if not self.send_query:
            return
        client = aioxmpp.PresenceManagedClient(aioxmpp.JID.fromstr(self.user), aioxmpp.make_security_layer(self.passwd))
        async with client.connected():
            while self.send_query:
                line = self.send_query.pop()
                message = aioxmpp.Message(to=aioxmpp.JID.fromstr(line.recepient), type_=aioxmpp.MessageType.CHAT)
                message.body[None] = line.message
                await client.send(message)

    async def run(self):
        while True:
            try:
                await self.asend_pending()
            except Exception:
                self.logger.exception(f'failed to send jabber messages')
                await asyncio.sleep(ON_ERROR_RETRY_DELAY)
            await asyncio.sleep(1)



from functools import wraps
from pathlib import Path
from typing import Any, Dict, Generic, List, Tuple, Type, TypeVar

from pydantic import BaseModel, ValidationError, create_model

from avtdl.core.chain import Chain, ChainConfigSection
from avtdl.core.loggers import LogLevel, override_loglevel, set_file_logger
from avtdl.core.plugins import Plugins


class ConfigurationError(Exception):
    '''Generic exception raised if parsing config failed'''

def format_validation_error(e: ValidationError) -> str:
    msg = 'Failed to process configuration file, following errors occurred: '
    errors = []
    for err in e.errors():
        location = ': '.join(str(l) for l in err['loc'])
        error = 'error parsing "{}" in config section "{}": {}'
        errors.append(error.format(err['input'], location, err['msg']))
    return '\n    '.join([msg] + errors)

def try_parsing(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ValidationError as e:
            error = format_validation_error(e)
            raise ConfigurationError(error) from e
    return wrapper

class SettingsSection(BaseModel):
    plugins_directory: str = 'plugins'
    log_directory: Path = Path('logs')
    logfile_size: int = 1000000
    logfile_level: LogLevel = LogLevel.debug
    loglevel_override: Dict[str, LogLevel] = {'bus': LogLevel.info, 'chain': LogLevel.info, 'actor.request': LogLevel.info}

def configure_loggers(settings: SettingsSection):
    override_loglevel(settings.loglevel_override)
    set_file_logger(path=settings.log_directory, max_size=settings.logfile_size, level=settings.logfile_level)

class ActorConfigSection(BaseModel):
    config: dict = {}
    defaults: dict = {}
    entities: List[dict]

class Config(BaseModel):
    Settings: SettingsSection = SettingsSection()
    Actors: Dict[str, ActorConfigSection]
    Chains: Dict[str, ChainConfigSection]


TConfig = TypeVar('TConfig')
TEntity = TypeVar('TEntity')

class SpecificActorConfigSection(BaseModel, Generic[TConfig, TEntity]):
    config: TConfig
    entities: List[TEntity]

class ActorParser:

    @staticmethod
    def flatten_actor_section(name: str, section: ActorConfigSection) -> ActorConfigSection:
        config = {**section.config, **{'name': name}}
        data: Dict[str, Any] = {'name': name, 'config': config, 'entities': []}
        for entity in section.entities:
            data['entities'].append({**section.defaults, **entity})
        return ActorConfigSection(**data)

    @staticmethod
    def load_actors_plugins_model(actor_section: dict) -> Dict[str, SpecificActorConfigSection]:
        actors_models: Dict[str, Any] = {}
        for name, section in actor_section.items():
            _, ConfigFactory, EntityFactory = Plugins.get_actor_factories(name)
            model = SpecificActorConfigSection[ConfigFactory, EntityFactory]
            actors_models[name] = (model, ...)
        actors_section_model = create_model('SpecificActors', **actors_models)
        return actors_section_model

    @classmethod
    def create_actors(cls, config_section: 'SpecificActors') -> Dict[str, Type]:
        actors = {}
        for name, actor_section in config_section:
            ActorFactory, _, _ = Plugins.get_actor_factories(name)
            actors[name] = ActorFactory(actor_section.config, actor_section.entities)
        return actors


class ConfigParser:

    @staticmethod
    def flatten_config(config: Config) -> Config:
        conf = config.model_dump()
        actors_section: Dict[str, ActorConfigSection] = {}
        for name, section in config.Actors.items():
            actors_section[name] = ActorParser.flatten_actor_section(name, section)
        conf['Actors'] = actors_section
        return Config(**conf)

    @staticmethod
    def load_models(config: Config) -> Type['SpecificConfig']:
        actors_model = ActorParser.load_actors_plugins_model(config.Actors)
        SpecificConfigModel = create_model('SpecificConfig',
                                     Actors=(actors_model, ...),
                                     Chains=(Dict[str, ChainConfigSection], ...)
                                     )
        return SpecificConfigModel

    @classmethod
    def create_chains(cls, chains_section: Dict[str, ChainConfigSection]) -> Dict[str, Chain]:
        chains = {}
        for name, chain_config in chains_section.items():
            chains[name] = Chain(name, chain_config)
        return chains

    @classmethod
    @try_parsing
    def parse(cls, conf) -> Tuple[Dict[str, Any], Dict[str, Chain]]:
        # do basic structural validation of config file
        config = Config(**conf)

        configure_loggers(config.Settings)
        Plugins.load(config.Settings.plugins_directory)

        # after that entities transformation and specific plugins validation can be safely performed
        flatted_conf = cls.flatten_config(config)
        SpecificConfig = cls.load_models(config)
        specific_config = SpecificConfig(**flatted_conf.model_dump())

        actors = ActorParser.create_actors(specific_config.Actors)
        chains = ConfigParser.create_chains(specific_config.Chains)

        return actors, chains





#!/usr/bin/env python3

from typing import Optional, Sequence

import aiohttp
from pydantic import Field

from avtdl.core.config import Plugins
from avtdl.core.interfaces import ActorConfig, TextRecord
from avtdl.core.monitors import HttpTaskMonitor, HttpTaskMonitorEntity

Plugins.register('get_url', Plugins.kind.ASSOCIATED_RECORD)(TextRecord)

@Plugins.register('get_url', Plugins.kind.ACTOR_CONFIG)
class UrlMonitorConfig(ActorConfig):
    pass

@Plugins.register('get_url', Plugins.kind.ACTOR_ENTITY)
class UrlMonitorEntity(HttpTaskMonitorEntity):
    url: str
    """url to monitor"""
    last_record_hash: Optional[str] = Field(exclude=True, default=None)
    """internal variable to persist state between updates. Used to keep hash of the page contents on previous update"""

@Plugins.register('get_url', Plugins.kind.ACTOR)
class UrlMonitor(HttpTaskMonitor):
    """
    Monitor web page text

    Download contents of web page at `url` and emit it as a `TextRecord`
    if it has changed since the last update. Intended for working with simple
    text endpoints.
    """

    async def get_new_records(self, entity: UrlMonitorEntity, session: aiohttp.ClientSession) -> Sequence[TextRecord]:
        text = await self.request(entity.url, entity, session)
        if text is None:
            return []
        record = TextRecord(text=text)
        record_hash = record.hash()
        if record_hash != entity.last_record_hash:
            entity.last_record_hash = record_hash
            return [record]
        return []



import logging
import sqlite3
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, Sequence, Union

import aiohttp
import feedparser
from pydantic import ConfigDict

from avtdl.core import utils
from avtdl.core.config import Plugins
from avtdl.core.db import BaseRecordDB
from avtdl.core.interfaces import Record
from avtdl.plugins.rss.generic_rss import GenericRSSMonitor, GenericRSSMonitorConfig, GenericRSSMonitorEntity
from avtdl.plugins.youtube import video_info
from avtdl.plugins.youtube.common import thumbnail_url


@Plugins.register('rss', Plugins.kind.ASSOCIATED_RECORD)
class YoutubeFeedRecord(Record):
    """
    Youtube video or livestream parsed from a channel's RSS feed
    """
    model_config = ConfigDict(extra='allow')

    url: str
    """link to the video"""
    title: str
    """title of the video at the time of parsing"""
    published: datetime
    """published value of the feed item, usually the time when the video was uploaded or the livestream frame was set up"""
    updated: datetime
    """updated value of the feed item. If different from `published`, might indicate either a change to video title, thumbnail or description, or a change in video status, for example livestream ending"""
    author: str
    """author's name, as shown on the channel icon"""
    video_id: str
    """short string identifying the video on Youtube. Part of the video url"""
    summary: str
    """video's description"""
    views: Optional[int]
    """current number of views. Is zero for upcoming and ongoing livestreams"""
    scheduled: Optional[datetime] = None
    """scheduled time for an upcoming livestream to start at, otherwise absent"""

    async def check_scheduled(self, session: Optional[aiohttp.ClientSession] = None):
        if self.views == 0:
            try:
                info = await video_info.aget_video_info(self.url, session)
                scheduled = info.scheduled
            except Exception:
                logging.exception('Exception while trying to get "scheduled" field, skipping')
                scheduled = None
        else:
            scheduled = None
        self.scheduled = scheduled

    def __str__(self):
        return self.format_record()

    def __repr__(self):
        template = '{} {:<8} [{}] {}'
        return template.format(self.format_date(self.published), self.author, self.video_id, self.title[:60])

    @staticmethod
    def format_date(date: Union[str, datetime]) -> str:
        if isinstance(date, str):
            date = datetime.fromisoformat(date)
        return date.strftime('%Y-%m-%d %H:%M')

    def format_record(self):
        scheduled = self.scheduled
        if scheduled:
            scheduled_time = '\nscheduled to {}'.format(self.format_date(scheduled))
        else:
            scheduled_time = ''
        template = '{}\n{}\npublished by {} at {}'
        return template.format(self.url, self.title, self.author, self.format_date(self.published)) + scheduled_time

    def as_dict(self, additional_fields):
        record_dict = {}
        record_dict.update(self.__dict__)
        record_dict.update(additional_fields)
        return record_dict

    def discord_embed(self) -> dict:
        embed = {
            'title': self.title,
            # 'description': ,
            'url': self.url,
            'color': None,
            'author': {'name': self.author},
            'timestamp': self.format_date(self.published),
            'image': {'url': thumbnail_url(self.video_id)}
        }
        if self.scheduled is not None:
            scheduled = self.scheduled.strftime('%Y-%m-%d %H:%M')
            embed['fields'] = [{'name': 'Scheduled:', 'value': scheduled, 'inline': True}]
        return embed


class RecordDB(BaseRecordDB):
    table_structure = 'parsed_at datetime, feed_name text, author text, video_id text, link text, title text, summary text, published datetime, updated datetime, scheduled datetime DEFAULT NULL, views integer, PRIMARY KEY(video_id, updated)'
    row_structure = ':parsed_at, :feed_name, :author, :video_id, :url, :title, :summary, :published, :updated, :scheduled, :views'
    id_field = 'video_id'
    exact_id_field = 'updated'
    group_id_field = 'feed_name'
    sorting_field = 'parsed_at'

    def store(self, rows: Union[Dict[str, Any], List[Dict[str, Any]]]) -> None:
        return super().store(rows)

    def row_exists(self, video_id: str, updated: Optional[datetime] = None) -> bool:
        return super().row_exists(video_id, updated)

    def fetch_row(self, video_id: str, updated: Optional[datetime] = None) -> Optional[sqlite3.Row]:
        return super().fetch_row(video_id, updated)

    def get_size(self, feed_name: Optional[str] = None) -> int:
        '''return number of records, total or for specified feed, are stored in db'''
        return super().get_size(feed_name)


@Plugins.register('rss', Plugins.kind.ACTOR_ENTITY)
class FeedMonitorEntity(GenericRSSMonitorEntity):
    update_interval : float = 900
    """How often the feed should be updated, in seconds"""
    track_reschedule: bool = True
    """keep track of scheduled time of upcoming streams, emit record again if it is changed to an earlier date"""

@Plugins.register('rss', Plugins.kind.ACTOR_CONFIG)
class FeedMonitorConfig(GenericRSSMonitorConfig):
    pass

@Plugins.register('rss', Plugins.kind.ACTOR)
class FeedMonitor(GenericRSSMonitor):
    """
    Youtube channel RSS feed monitor

    Monitors channel for new uploads and livestreams using RSS feed
    generated by Youtube. Requires old channel id format. In order to obtain
    RSS feed url for a given channel, use "View Source" on a channel page
    and search for "rss".

    Example of a supported url:

    - `https://www.youtube.com/feeds/videos.xml?channel_id=UCK0V3b23uJyU4N8eR_BR0QA`

    RSS feed is smaller and faster to parse compared to an HTML channel page,
    but by design only shows updates of a single channel and doesn't support
    authentication and therefore unable to show member-only streams.

    Scheduled date for upcoming streams is not present in the feed itself, so it
    is obtained by fetching and parsing video page the first time it appears in
    the feed. It then gets updated until stream goes live, unless `track_reschedule`
    option is disabled.

    No matter how often the url gets fetched, content of the feed only gets
    changed once every 15 minutes, so setting `update_interval` lower than that
    value is not recommended. This monitor will attempt to calculate time of the
    next update from HTTP headers and schedule next request right after it. Use
    `adjust_update_interval` to disable this behavior.
    """

    RecordDB = RecordDB

    def __init__(self, conf: FeedMonitorConfig, entities: Sequence[FeedMonitorEntity]):
        super().__init__(conf, entities)

    async def get_new_records(self, entity: FeedMonitorEntity, session: aiohttp.ClientSession) -> Sequence[YoutubeFeedRecord]:
        records = await self.get_records(entity, session)
        new_records = []

        # record.check_scheduled() involves loading video page, so it should only be done when necessarily
        # and here seems to be the only good place for it, since network request requires "session" object
        for record in records:
            previous = self.load_record(record, entity)
            if previous is None:
                await record.check_scheduled(session)
                continue
            if not entity.track_reschedule:
                continue
            if previous is not None and previous.scheduled is not None:
                self.logger.debug(f'{record.video_id=} has last {previous.scheduled=}, updating')
                await record.check_scheduled(session)
                if record.scheduled is None:
                    continue
                if record.scheduled < previous.scheduled:
                    # treat rescheduled records as new if scheduled time is earlier than before
                    # to allow action run on time, though it will run second time later
                    msg = 'In feed {} record [{}] {} rescheduled back from {} to {}'
                    msg = msg.format(entity.name, record.video_id, record.title, previous.scheduled, record.scheduled)
                    self.logger.warning(msg)
                    new_records.append(record)

        new_records.extend(self.filter_new_records(records, entity))
        return new_records

    def get_record_id(self, record: YoutubeFeedRecord) -> str:
        return record.video_id

    def _get_record_id(self, record: YoutubeFeedRecord, entity: FeedMonitorEntity) -> str:
        return self.get_record_id(record)

    def record_got_updated(self, record: YoutubeFeedRecord, entity: FeedMonitorEntity) -> bool:
        return self.db.row_exists(record.video_id) and not self.db.row_exists(record.video_id, record.updated)

    def store_records(self, records: Sequence[YoutubeFeedRecord], entity: FeedMonitorEntity):
        rows = []
        for record in records:
            now = datetime.now(tz=timezone.utc).isoformat(timespec='seconds')
            additional_fields = {'feed_name': entity.name, 'parsed_at': now}
            row = record.as_dict(additional_fields)
            rows.append(row)
        self.db.store(rows)

    def load_record(self, record: YoutubeFeedRecord, entity: FeedMonitorEntity) -> Optional[YoutubeFeedRecord]:
        raw_latest_row = self.db.fetch_row(record.video_id)
        if raw_latest_row is None:
            return None
        latest_row = dict(raw_latest_row)
        for field in ['parsed_at', 'feed_name']:
            latest_row.pop(field)
        latest_row['url'] = latest_row.pop('link')
        for field in ['published', 'updated', 'scheduled']:
            value = latest_row[field]
            # handle db records with old format
            if isinstance(value, str):
                latest_row[field] = datetime.fromisoformat(value)
        return YoutubeFeedRecord(**latest_row)

    @classmethod
    def _parse_entry(cls, entry: feedparser.FeedParserDict) -> YoutubeFeedRecord:
        parsed = {}
        parsed['url'] = entry['link']
        parsed['title'] = entry['title']
        parsed['updated'] = utils.make_datetime(entry['updated_parsed'])

        parsed['published'] = utils.make_datetime(entry['published_parsed'])
        parsed['author'] = entry.get('author')
        parsed['summary'] = entry.get('summary')
        parsed['video_id'] = entry.get('yt_videoid', 'video_id missing')
        try:
            views = int(entry['media_statistics']['views'])
        except (ValueError, KeyError, TypeError):
            views = None
        parsed['views'] = views
        record = YoutubeFeedRecord(**parsed)
        return record



import asyncio
import datetime
import json
import logging
from textwrap import shorten
from typing import Any, List, Optional, Sequence, Tuple

import aiohttp
import dateutil
import multidict
from pydantic import Field, field_validator

from avtdl.core.interfaces import Actor, ActorConfig, ActorEntity, Record
from avtdl.core.plugins import Plugins

EMBEDS_PER_MESSAGE = 10
EMBED_TITLE_MAX_LENGTH = 256
EMBED_DESCRIPTION_MAX_LENGTH = 4096


class DiscordWebhook:

    def __init__(self, name: str, hook_url: str, logger: Optional[logging.Logger] = None):
        self.logger = logger or logging.getLogger('DiscordHook')
        self.name = name
        self.hook_url = hook_url
        self.send_query: Optional[asyncio.Queue] = None
        self.session: Optional[aiohttp.ClientSession] = None

    def to_be_sent(self, record: Record):
        if self.send_query is None:
            return
        self.logger.debug(f'adding record to send query: {record!r}')
        self.send_query.put_nowait(record)

    async def run(self) -> None:
        self.send_query = asyncio.Queue()
        self.session = aiohttp.ClientSession()

        until_next_try = 0
        to_be_sent: List[Record] = []
        while True:
            await asyncio.sleep(until_next_try)
            try:
                while len(to_be_sent) < EMBEDS_PER_MESSAGE:
                    record = await asyncio.wait_for(self.send_query.get(), 60/EMBEDS_PER_MESSAGE)
                    to_be_sent.append(record)
            except asyncio.TimeoutError:
                pass
            if len(to_be_sent) == 0:
                continue
            message, pending_records = MessageFormatter.format(to_be_sent)
            try:
                limits_ok = MessageFormatter.check_limits(message)
            except Exception as e:
                self.logger.debug(f'error checking content length limits for message: {e}\nRaw message: "{message}"')
                continue
            if not limits_ok:
                self.logger.warning(f'[{self.name}] prepared message exceeded Discord length limits. Records will be discarded')
                self.logger.debug(f'[{self.name}] message content:\n{message}')
                to_be_sent = pending_records
                continue
            try:
                response = await self.session.post(self.hook_url, json=message)
                text = await response.text()
            except OSError as e:
                self.logger.warning(f'[{self.name}] error while sending message with Discord webhook: {e}')
                until_next_try = 60
                continue  # implicitly saving messages in to_be_sent until the next try
            except Exception as e:
                self.logger.exception(f'[{self.name}] error while sending message with Discord webhook: {e}')
                self.logger.debug(f'raw message text:\n{message}')
                to_be_sent = pending_records
                until_next_try = 60
                continue
            try:
                response.raise_for_status()
            except aiohttp.ClientResponseError as e:
                self.logger.warning(f'got {response.status} ({response.reason}) from {self.hook_url}')
                if e.status == 400:
                    self.logger.warning(f'[{self.name}] message got rejected with {response.status} ({response.reason}), dropping it')
                    self.logger.debug(f'[{self.name}] response headers: {response.headers}')
                    self.logger.debug(f'[{self.name}] raw request body: {json.dumps(message)}')
                elif e.status in  [404, 403]:
                    self.logger.warning(f'[{self.name}] got {e.status} from webhook, interrupting operations. Check if webhook url is still valid: {self.hook_url}')
                    break
            # if message got send successfully discard records except these that didn't fit into message
            to_be_sent = pending_records
            until_next_try = self.get_next_delay(response.headers)

    def get_next_delay(self, headers: multidict.CIMultiDictProxy):
        if 'Retry-After' in headers:
            delay = headers.get('Retry-After', 0)
            self.logger.debug(f'Retry-After header is set to {delay}')
        else:
            remaining = headers.get('X-RateLimit-Remaining', '0')
            self.logger.debug(f'X-RateLimit-Remaining={remaining}')
            if remaining == '0':
                delay = headers.get('X-RateLimit-Reset-After', 'no X-RateLimit-Reset-After header in response')
                self.logger.debug(f'X-RateLimit-Reset-After={delay}')
            else:
                delay = 0
        try:
            delay = int(delay)
        except ValueError:
            self.logger.debug(f'failed to parse delay {delay} for {self.hook_url}, using default')
            delay = 6
        return delay


class MessageFormatter:

    @classmethod
    def format(cls, records: List[Record]) -> Tuple[dict, List[Record]]:
        '''take records and format them in Discord webhook payload as embeds
        after the limit on embeds is reached, the rest of the records are returned back'''
        embeds: List[dict] = []
        excess_records = []
        for i, record in enumerate(records):
            record_embeds = cls.make_embeds(record)
            if len(embeds) + len(record_embeds) > EMBEDS_PER_MESSAGE:
                excess_records = records[i:]
                break
            else:
                embeds.extend(record_embeds)
        message = cls.make_message(embeds)
        return message, excess_records

    @classmethod
    def make_message(cls, embeds: List[dict]) -> dict:
        return {
            "content": None,
            "embeds": embeds
        }

    @classmethod
    def make_embeds(cls, record: Record) -> List[dict]:
        formatter = getattr(record, 'discord_embed', None)
        if formatter is not None and callable(formatter):
            embeds = formatter()
            if not isinstance(embeds, list):
                embeds = [embeds]
            return embeds
        return [cls.plaintext_embed(record)]

    @classmethod
    def plaintext_embed(cls, record: Record) -> dict:
        text = str(record)
        if text.find('\n') > -1:
            title, description = text.split('\n', 1)
        else:
            title, description = '', text
        title = shorten(title, EMBED_TITLE_MAX_LENGTH)
        description = shorten(description, EMBED_DESCRIPTION_MAX_LENGTH)
        return {'title': title, 'description': description}

    @classmethod
    def check_limits(cls, message: dict) -> bool:
        # doesn't count field.name and field.value number and size in hope it will not change outcome

        class Limits:
            TOTAL = 6000
            AUTHOR_NAME = 256
            TITLE = 256
            DESCRIPTION = 4096
            FOOTER_TEXT = 2048

        total_length = 0
        embeds = message.get('embeds', [])
        for embed in embeds:
            author_name = len(embed.get('author', {}).get('name', ''))
            title = len(embed.get('title') or '')
            description = len(embed.get('description') or '')
            footer_text = len((embed.get('footer') or {}).get('text') or '')
            if author_name > Limits.AUTHOR_NAME:
                return False
            if title > Limits.TITLE:
                return False
            if description > Limits.DESCRIPTION:
                return False
            if footer_text > Limits.FOOTER_TEXT:
                return False
            total_length += author_name + title + description + footer_text

        if total_length > Limits.TOTAL:
            return False
        return True


@Plugins.register('discord.hook', Plugins.kind.ACTOR_CONFIG)
class DiscordHookConfig(ActorConfig):
    pass


@Plugins.register('discord.hook', Plugins.kind.ACTOR_ENTITY)
class DiscordHookEntity(ActorEntity):
    url: str
    """webhook url"""
    timezone: Optional[str] = None # https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
    """takes timezone name from <https://en.wikipedia.org/wiki/List_of_tz_database_time_zones> (or OS settings if omitted), converts record fields containing date and time to this timezone"""
    hook: Optional[Any] = Field(exclude=True, default=None)
    """internal variable to persist state between update calls, holds DiscordWebhook object for this entity"""

    @field_validator('timezone')
    @classmethod
    def check_timezone(cls, timezone: str) -> datetime.timezone:
        tz = dateutil.tz.gettz(timezone)
        if tz is None:
            raise ValueError(f'Unknown timezone: {timezone}')
        return tz

@Plugins.register('discord.hook', Plugins.kind.ACTOR)
class DiscordHook(Actor):
    """
    Send record to Discord using webhook

    To generate webhook url follow instructions in "Making a Webhook" section of
    <https://support.discord.com/hc/en-us/articles/228383668-Intro-to-Webhooks>

    Some record types support rich formatting when sent to Discord, such as
    showing the author's avatar and links to attached images. Youtube videos will
    show thumbnail, however embedding video itself is not supported.

    Records coming within six seconds one after another will be batched together into a single message.
    When too many records are received at once, they will be sent with delays to conform to Discord
    rate limits. Records deemed to be too long to fit in a Discord message
    [length limits](https://discord.com/developers/docs/resources/channel#create-message-jsonform-params)
    will be dropped with a warning.
    """

    def __init__(self, conf: DiscordHookConfig, entities: Sequence[DiscordHookEntity]):
        super().__init__(conf, entities)
        for entity in entities:
            entity.hook = DiscordWebhook(entity.name, entity.url, self.logger)

    def handle(self, entity: DiscordHookEntity, record: Record):
        if entity.hook is None:
            return
        entity.hook.to_be_sent(record.as_timezone(entity.timezone))

    async def run(self):
        tasks = [asyncio.create_task(entity.hook.run()) for entity in self.entities.values()]
        await asyncio.wait(tasks)
